from contextlib import contextmanager
import warnings

from .logging import log

class AICertException(Exception):
    """Generic AICert exception"""
    def __init__(self, message: str, *args: object) -> None:
        self.message = message
        super().__init__(self.message, *args)

@contextmanager
def log_errors_and_warnings():
    """Context manager that intercepts AICert errors and warnings and that logs them
    
    Only errors that inherit from AICertClientException are caught.
    They are used to display a nice error message before cleanly exiting the program.
    All warnings are caught and logged (and the program continues).

    Example usage:
    ```py
    with log_errors_and_warnings():
        # Potential errors are logged
        # and the program is properly terminated
        function_that_may_fail()
    ```
    """
    try:
        yield None
    except AICertException as e:
        log.error(f"{e.message}")
        exit(1)
    finally:
        with warnings.catch_warnings(record=True) as ws:
            for w in ws:
                log.warning(w.message)

class ValidationError(Exception):
    def __init__(self, message: str):
        super().__init__(message)

class APIKeyException(Exception):
    """
    A custom exception class for wrapping API Key verification errors
    """

    def __init__(self, key, msg):
        self.key = key
        self.msg = msg

    def __str__(self):
        return f"{self.msg}"
    
class PredictionException(Exception):
    """
    A custom exception class for wrapping predictions errors
    """

    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return f"{self.msg}"


import os
import logging
from rich.logging import RichHandler


LOG_LEVEL = os.getenv("AICERT_LOG_LEVEL")
if LOG_LEVEL is None or LOG_LEVEL not in [
    "CRITICAL",
    "ERROR",
    "WARNING",
    "INFO",
    "DEBUG",
]:
    LOG_LEVEL = "INFO"


logging.basicConfig(
    level=LOG_LEVEL,
    format="%(message)s",
    datefmt="[%X]",
    handlers=[RichHandler(rich_tracebacks=True)],
)

log = logging.getLogger("rich")



import base64
import hashlib
import json
import subprocess
import tempfile
import OpenSSL
from OpenSSL import crypto
import requests
import yaml
from .logging import log


from cryptography.hazmat.primitives import serialization
from cryptography.x509 import load_der_x509_certificate

PCR_FOR_CERTIFICATE = 15
PCR_FOR_MEASUREMENT = 16


class AttestationError(Exception):
    """This exception is raised when the attestation is invalid (enclave
    settings mismatching, debug mode unallowed...).

    Used as base exception for all other sub exceptions on the attestation
    validation
    """

    pass


def verify_ak_cert(cert_chain: list[bytes]) -> bytes:
    """
    Verify the certificate chain of the attestation key.
    Parameters:
        cert_chain: list of certificates in DER format
    Returns:
        AK certificate in DER format
    Raises:
        AttestationError: if the certificate chain is invalid
    """

    # Load certificate to be verified : attestation key certificate
    ak_cert = crypto.load_certificate(crypto.FILETYPE_ASN1, cert_chain[0])

    # Verify the certificate's chain
    store = crypto.X509Store()

    # Create the CA cert object from PEM string, and store into X509Store
    req = requests.get("http://crl.microsoft.com/pkiinfra/certs/AMERoot_ameroot.crt")
    req.raise_for_status()
    _rootca_cert = crypto.load_certificate(crypto.FILETYPE_ASN1, req.content)  # type: ignore
    store.add_cert(_rootca_cert)

    chain = [
        crypto.load_certificate(crypto.FILETYPE_ASN1, _cert_der)
        for _cert_der in cert_chain[1:-1]
    ]

    store_ctx = crypto.X509StoreContext(store, ak_cert, chain=chain)

    # try:
    #     # if the cert is invalid, it will raise a X509StoreContextError
    #     store_ctx.verify_certificate()
    # except crypto.X509StoreContextError:
    #     ref_cert = OpenSSL.crypto.load_certificate(OpenSSL.crypto.FILETYPE_ASN1,cert_chain[0])
    #     log.info(OpenSSL.crypto.dump_certificate(OpenSSL.crypto.FILETYPE_TEXT, ref_cert).decode('ascii'))
    #     ref_cert = OpenSSL.crypto.load_certificate(OpenSSL.crypto.FILETYPE_ASN1,cert_chain[1])
    #     log.info(OpenSSL.crypto.dump_certificate(OpenSSL.crypto.FILETYPE_TEXT, ref_cert).decode('ascii'))
    #     ref_cert = OpenSSL.crypto.load_certificate(OpenSSL.crypto.FILETYPE_ASN1,cert_chain[2])
    #     log.info(OpenSSL.crypto.dump_certificate(OpenSSL.crypto.FILETYPE_TEXT, ref_cert).decode('ascii'))
    #     log.error("Invalid AK certificate")
    #     # TODO: In the future we'll want to raise an error
    #     # But for now Azure vTPM does not have an official certificate chain
    #     # So the checks can randomly fail
    #     # raise AttestationError("Invalid AK certificate")

    return cert_chain[0]


def check_quote(quote, pub_key_pem):
    """
    Check quote using tpm2_checkquote command.
    Parameters:
         quote: dictionary with keys 'message', 'signature', and 'pcr'
         pub_key_pem: public key in PEM format (string)
    Returns:
    Raises:


    """
    with tempfile.NamedTemporaryFile() as quote_msg_file, tempfile.NamedTemporaryFile() as quote_sig_file, tempfile.NamedTemporaryFile() as quote_pcr_file, tempfile.NamedTemporaryFile() as ak_pub_key_file:
        quote_msg_file.write(quote["message"])
        quote_msg_file.flush()

        quote_sig_file.write(quote["signature"])
        quote_sig_file.flush()

        quote_pcr_file.write(quote["pcr"])
        quote_pcr_file.flush()

        ak_pub_key_file.write(pub_key_pem)
        ak_pub_key_file.flush()

        tpm2_checkquote = subprocess.run(
            [
                "tpm2_checkquote",
                "--public",
                ak_pub_key_file.name,
                "--message",
                quote_msg_file.name,
                "--pcr",
                quote_pcr_file.name,
                "--signature",
                quote_sig_file.name,
            ],
            check=True,
            capture_output=True,
            text=True,
        )
        att_document = yaml.load(tpm2_checkquote.stdout, Loader=yaml.BaseLoader)
        att_document["pcrs"]["sha256"] = {
            int(k): v.lower().removeprefix("0x")
            for k, v in att_document["pcrs"]["sha256"].items()
        }
        return att_document


def check_server_cert(
    received_cert,
    pcr_end,
    initial_pcr="0000000000000000000000000000000000000000000000000000000000000000",
):
    initial_pcr = bytes.fromhex(initial_pcr)
    current_pcr = initial_pcr

    hash_event = hashlib.sha256(received_cert.encode()).digest()
    current_pcr = hashlib.sha256(current_pcr + hash_event).digest()

    # log.info(f"Certificate PCR in quote : {pcr_end}")
    # log.info(f"Expected Certificate PCR : {current_pcr.hex()}")
    # Both PCR MUST match, else something sketchy is going on!
    if pcr_end != current_pcr.hex():
        return False
    
    return True


def check_event_log(
    input_event_log,
    pcr_end,
    initial_pcr="0000000000000000000000000000000000000000000000000000000000000000",
):
    # Starting from the expected initial PCR state
    # We replay the event extending the PCR
    # At the end we get the expected PCR value
    initial_pcr = bytes.fromhex(initial_pcr)
    current_pcr = initial_pcr
    for e in input_event_log:
        hash_event = hashlib.sha256(e.encode()).digest()
        current_pcr = hashlib.sha256(current_pcr + hash_event).digest()

    # log.info(f"Event PCR in quote : {pcr_end}")
    # log.info(f"Expected Event PCR : {current_pcr.hex()}")

    # Both PCR MUST match, else something sketchy is going on!
    assert pcr_end == current_pcr.hex()

    # Now we can return the parsed event log
    event_log = [json.loads(e) for e in input_event_log]

    return event_log


def decode_b64_encoding(x):
    return base64.b64decode(x["base64"])


def test_check_pass():
    with open("sample_build_response.json") as f:
        build_response = json.load(f)

    build_response["remote_attestation"]["cert_chain"] = [
        decode_b64_encoding(cert_b64_encoded)
        for cert_b64_encoded in build_response["remote_attestation"]["cert_chain"]
    ]

    ak_cert = verify_ak_cert(
        cert_chain=build_response["remote_attestation"]["cert_chain"]
    )

    ak_cert_ = load_der_x509_certificate(ak_cert)
    ak_pub_key = ak_cert_.public_key()
    ak_pub_key_pem = ak_pub_key.public_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PublicFormat.SubjectPublicKeyInfo,
    )

    build_response["remote_attestation"]["quote"] = {
        k: decode_b64_encoding(v)
        for k, v in build_response["remote_attestation"]["quote"].items()
    }
    att_document = check_quote(
        build_response["remote_attestation"]["quote"], ak_pub_key_pem
    )

    print("attestation document", att_document)

    # We should check the PCR to make sure the system has booted properly
    # This is an example ... the real thing will depend on the system.
    assert (
        att_document["pcrs"]["sha256"][0]
        == "d0d725f21ba5d701952888bcbc598e6dcef9aff4d1e03bb3606eb75368bab351"
    )
    assert (
        att_document["pcrs"]["sha256"][1]
        == "fe72566c7f411900f7fa1b512dac0627a4cac8c0cb702f38919ad8c415ca47fc"
    )
    assert (
        att_document["pcrs"]["sha256"][2]
        == "3d458cfe55cc03ea1f443f1562beec8df51c75e14a9fcf9a7234a13f198e7969"
    )
    assert (
        att_document["pcrs"]["sha256"][3]
        == "3d458cfe55cc03ea1f443f1562beec8df51c75e14a9fcf9a7234a13f198e7969"
    )
    assert (
        att_document["pcrs"]["sha256"][4]
        == "1f0105624ab37b9af59da6618a406860e33ef6f42a38ddaf6abfab8f23802755"
    )
    assert (
        att_document["pcrs"]["sha256"][5]
        == "d36183a4ce9f539d686160695040237da50e4ad80600607f84eff41cf394dcd8"
    )

    # To make test easier we use the PCR 16 since it is resettable `tpm2_pcrreset 16`
    # But because it is resettable it MUST NOT be used in practice.
    # An unused PCR that cannot be reset (SRTM) MUST be used instead
    # PCR 14 or 15 should do it
    print(
        check_event_log(
            build_response["event_log"],
            att_document["pcrs"]["sha256"][PCR_FOR_MEASUREMENT],
        )
    )



from typing import Optional, List
from pydantic import BaseModel, validator
from .errors import *
from .request_adapters import ForcedIPHTTPSAdapter
from .verify import *

import os
import requests
import json
import grpc
import blind_llama
import tempfile
import typer
import warnings

from grpc import ssl_channel_credentials

import blind_llama.pb.licensing_pb2_grpc as licensing_pb2_grpc
import blind_llama.pb.licensing_pb2 as licensing_pb2

MITHRIL_SERVICES_URL = os.getenv("MITHRIL_SERVICES_URL", "api.cloud.mithrilsecurity.io")

PCR_FOR_MEASUREMENT = 16
PCR_FOR_CERTIFICATE = 15

class AICertInvalidAttestationFormatException(AICertException):
    """AICert attestation parsing error (json)"""
    def __init__(self, err: Exception) -> None:
        self.__err = err
        self.message = f"Invalid attestation format\n{self.__err}"
        super().__init__(self.message)


class AICertInvalidAttestationException(AICertException):
    """Invalid attestation error"""
    pass


class PromptRequest(BaseModel):
    inputs: str
    temperature: float

    @validator("inputs")
    def valid_input(cls, v):
        if not v:
            raise ValidationError("The prompt cannot be empty")
        return v
    
    @validator("temperature")
    def valid_temperature(cls, v):
        if v < 0:
            raise ValidationError("Invalid temperature")
        return v

class Client():
    """A class to represent a connection to a BlindLlama server."""

    def __init__(
        self,
        verbose: bool = False
    ):
        self.verbose = verbose
        #self.__base_url = f"{openai2.api_base}:8000"
        self.__base_url = "https://llama_worker"
        self.__attest_url = "https://aicert_worker"
        self.__session = requests.Session()
        self.__session.mount(
            self.__base_url, ForcedIPHTTPSAdapter(dest_ip=blind_llama.api_base)
        )

        ca_cert = self.verify_server_certificate(blind_llama.security_config["expected_pcrs"]["codebasepcr"])

        server_ca_crt_file = tempfile.NamedTemporaryFile(mode="w+t", delete=False)
        server_ca_crt_file.write(ca_cert)
        server_ca_crt_file.flush()
        self.__session.verify = server_ca_crt_file.name

        channel = grpc.secure_channel(
                MITHRIL_SERVICES_URL,
                ssl_channel_credentials(),
            )
        
        api_key_env = blind_llama.api_key
        stub = licensing_pb2_grpc.LicensingServiceStub(channel)
        try:
            enclave_request = licensing_pb2.GetEnclaveRequest(api_key=api_key_env)
            response = stub.GetEnclave(enclave_request)
        except grpc.RpcError as rpc_error:
            raise APIKeyException(api_key_env, rpc_error.details())

        if len(response.jwt) > 0: 
            self.jwt = response.jwt
        else:
            raise APIKeyException(
                api_key_env, "The API Key you provided is invalid."
        )

        super(Client, self).__init__()

    def verify_server_certificate(self, expected_pcr):
        """Retrieve server certificate and validate it with 
        the attestation report.
        """
        session = requests.Session()
        session.mount(
                self.__attest_url, ForcedIPHTTPSAdapter(dest_ip=blind_llama.api_base)
            )
        
        attestation = session.get(f"{self.__attest_url}/aTLS",verify=False)

        attestation_json = json.loads(attestation.content)
        ca_cert = attestation_json["ca_cert"]

        # Verify quote and CA TLS certificate
        self.verify_build_response(attestation.content, ca_cert, expected_pcr)
        return ca_cert
    
    def verify_build_response(self, build_response: bytes, ca_cert = "", expected_pcr = ""):
        """Verify received attesation validity

        1. Parse the JSON reponse
        2. Check simulation mode
        3. Verify certificate chain
        4. Verify quote signature
        5. Verify boot PCRs (firmware, bootloader, initramfs, OS)
        6. Verify event log (final hash in PCR_FOR_MEASUREMENT) by replaying it (works like a chain of hashes)
        7. Verify TLS certificate (final hash in PCR_FOR_CERTIFICATE)
        
        Args:
            build_response (bytes): reponse of the attestation endpoint
            verbose (bool, default = False): whether to print verification information in stdout
        """
        try:
            build_response = json.loads(build_response)
        except Exception as e:
            AICertInvalidAttestationFormatException(e)
        
        if "simulation_mode" in build_response["remote_attestation"]:
            if self.__simulation_mode:
                warnings.warn(f"Attestation generated in simulation mode", RuntimeWarning)
                return
            else:
                raise AICertInvalidAttestationException(f"Attestation generated in simulation mode")

        build_response["remote_attestation"]["cert_chain"] = [
            decode_b64_encoding(cert_b64_encoded)
            for cert_b64_encoded in build_response["remote_attestation"]["cert_chain"]
        ]

        ak_cert = verify_ak_cert(
            cert_chain=build_response["remote_attestation"]["cert_chain"]
        )
        if self.verbose:
            warnings.warn(f"Bypassing certificate chain verification", RuntimeWarning)

        ak_cert_ = load_der_x509_certificate(ak_cert)
        ak_pub_key = ak_cert_.public_key()
        ak_pub_key_pem = ak_pub_key.public_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PublicFormat.SubjectPublicKeyInfo,
        )

        build_response["remote_attestation"]["quote"] = {
            k: decode_b64_encoding(v)
            for k, v in build_response["remote_attestation"]["quote"].items()
        }
        att_document = check_quote(
            build_response["remote_attestation"]["quote"], ak_pub_key_pem
        )

        if self.verbose:
            typer.secho(f"Valid quote", fg=typer.colors.GREEN)

            log.info(
                f"Attestation Document > PCRs :  \n{yaml.safe_dump(att_document['pcrs']['sha256'])}"
            )

        # To make test easier we use the PCR 16 since it is resettable `tpm2_pcrreset 16`
        # But because it is resettable it MUST NOT be used in practice.
        # An unused PCR that cannot be reset (SRTM) MUST be used instead
        # PCR 14 or 15 should do it
        event_log = check_event_log(
            build_response["event_log"],
            att_document["pcrs"]["sha256"][PCR_FOR_MEASUREMENT],
        )

        measured_pcr = att_document["pcrs"]["sha256"][PCR_FOR_MEASUREMENT]

        if self.verbose:
            log.info(f"Server's Codebase PCR : {measured_pcr}")
            log.info(f"Expected Codebase PCR : {expected_pcr}")

        if expected_pcr == measured_pcr:
            if self.verbose:
                typer.secho(f"Code base PCR is expected value")
        else:
            if self.verbose:
                typer.secho(f"Codebase PCR does not match expected value")
            raise ValidationError(f"Expected Codebase PCR does not match the one provided by the server. Expected PCR: {expected_pcr} Measured PCR: {measured_pcr}")


        result = check_server_cert(
            ca_cert,
            att_document["pcrs"]["sha256"][PCR_FOR_CERTIFICATE],
        )
        if not result:
            # Disconnect destroys the runner, this might not be required for an attestation failure
            # self.disconnect()
            raise AICertInvalidAttestationException(f"Attestation validation failed.")
        
        if self.verbose:
            typer.secho(f"Valid event log")
            print(yaml.safe_dump(event_log))
            typer.secho(f"ALL CHECKS PASSED")

        warnings.warn(f"The quote from the TPM is not endorsed by the Cloud provider for the alpha version of BlindLlama v0.1. For more information look at https://github.com/mithril-security/blind_llama")

    def predict(
        self,
        prompt: str,
        temperature: float,
    ) -> str:
        """Start a prediction.
        Args:
            prompt (str): The prompt on which you want to run a prediction on.
            temperature (float): The temperature requested.
        Returns:
            str: The result of the prediction made by the server
        """

        req = PromptRequest(inputs=prompt, temperature=temperature)
        Headers = { "accesstoken" : self.jwt }
        resp = self.__session.post(
            self.__base_url + "/",
            json=req.dict(),
            headers=Headers
        )
        ret_json = resp.json()
        generated_text = ""
        for elem in ret_json:
            if "error" in elem:
                raise PredictionException(elem)
            if "generated_text" in elem:
                generated_text += f"{elem['generated_text'].strip()}\n"
        return generated_text.strip()

def create(model: str = "meta-llama/Llama-2-70b-chat-hf", prompt: str = "", temperature: float = 0.7, verbose: bool = False) -> str:
    """Start a prediction.
    Args:
        model (str): The model on which you want to run a prediction on (default to meta-llama/Llama-2-70b-chat-hf).
        prompt (str): The prompt on which you want to run a prediction on.
        temperature (float): The temperature requested (default to 0.7).
    Returns:
        str: The result of the prediction made by the server
    """
    return Client(verbose).predict(prompt, temperature)



import os

__all__ = []

try:
    from . import completion
    from .completion import Client, create

    api_key = os.getenv("BLIND_LLAMA_API_KEY", "")
    api_base = "137.116.212.25"
    security_config = {
        "description": "Base config of BlindLlama",
        "expected_pcrs":
        {
            "codebasepcr": "a52bdc8b0fde6ba40ce06532a303e035d55baaf5e657d225093d54b6e05433bc",
            "tpmcert_pcr": "306f9d8b94f17d93dc6e7cf8f5c79d652eb4c6c4d13de2dddc24af416e13ecaf"
        },
        "source_binary": "ghcr.io/mithril-security/text-generation-inference:sha256-5c5f1fbaae44e760fe46c792d8bff31d52cd572c6ade394617f7867b7400cd2d.att",
        "source_code": "https://github.com/mithril-security/text-generation-inference",
        "audit_certificates":
        {
            ""
        }
    }

    __all__ += ["completion", "Client", "create", "api_base", "api_key", "security_config"]

except ImportError as e:
    print(e)
    pass
